"""Tests config Settings class."""
import os
import os.path
import sys
from pathlib import Path
from typing import Dict, List, Optional

import toml
from pydantic import BaseSettings, ValidationError

SETTINGS = None


class Settings(BaseSettings):  # pylint: disable=too-few-public-methods
    """Main Settings Class for the project.

    The type of each setting is defined using Python annotations
    and is validated when a config file is loaded with Pydantic.

    Most input files specific to this project are expected to be located in the same directory. e.g.
    schema/
     - definitions
     - schemas
    """

    # Main directory names
    main_directory: str = "schema"
    definition_directory: str = "definitions"
    schema_directory: str = "schemas"
    validator_directory: str = "validators"
    test_directory: str = "tests"

    # Settings specific to the schema files
    schema_file_extensions: List[str] = [".json", ".yaml", ".yml"]  # Do we still need that ?
    schema_file_exclude_filenames: List[str] = []

    # settings specific to search and identify all instance file to validate
    data_file_search_directories: List[str] = ["./"]
    data_file_extensions: List[str] = [".json", ".yaml", ".yml"]
    data_file_exclude_filenames: List[str] = [".yamllint.yml", ".travis.yml"]
    data_file_automap: bool = True

    ansible_inventory: Optional[str]
    schema_mapping: Dict = {}

    class Config:  # pylint: disable=too-few-public-methods
        """Additional parameters to automatically map environment variable to some settings."""

        fields = {
            "main_directory": {"env": "jsonschema_directory"},
            "definition_directory": {"env": "jsonschema_definition_directory"},
        }


def load(config_file_name="pyproject.toml", config_data=None):
    """Load configuration.

    Configuration is loaded from a file in pyproject.toml format that contains the settings,
    or from a dictionary of those settings passed in as "config_data"

    The settings for this app are expected to be in [tool.json_schema_testing] in TOML
    if nothing is found in the config file or if the config file do not exist, the default values will be used.

    config_data can be passed in to override the config_file_name. If this is done, a combination of the data
    specified and the defaults for parameters not specified will be used, and settings in the config file will
    be ignored.

    Args:
        config_file_name (str, optional): Name of the configuration file to load. Defaults to "pyproject.toml".
        config_data (dict, optional): dict to load as the config file instead of reading the file. Defaults to None.
    """
    global SETTINGS  # pylint: disable=global-statement

    if config_data:
        SETTINGS = Settings(**config_data)
        return
    if os.path.exists(config_file_name):
        config_string = Path(config_file_name).read_text(encoding="utf-8")
        config_tmp = toml.loads(config_string)

        if "tool" in config_tmp and "schema_enforcer" in config_tmp.get("tool", {}):
            SETTINGS = Settings(**config_tmp["tool"]["schema_enforcer"])
            return

    SETTINGS = Settings()


def load_and_exit(config_file_name="pyproject.toml", config_data=None):
    """Calls load, but wraps it in a try except block.

    This is done to handle a ValidationErorr which is raised when settings are specified but invalid.
    In such cases, a message is printed to the screen indicating the settings which don't pass validation.

    Args:
        config_file_name (str, optional): [description]. Defaults to "pyproject.toml".
        config_data (dict, optional): [description]. Defaults to None.
    """
    try:
        load(config_file_name=config_file_name, config_data=config_data)
    except ValidationError as err:
        print(f"Configuration not valid, found {len(err.errors())} error(s)")
        for error in err.errors():
            print(f"  {'/'.join(error['loc'])} | {error['msg']} ({error['type']})")
        sys.exit(1)



"""Validation related classes."""
from typing import List, Optional, Any
from pydantic import BaseModel, validator  # pylint: disable=no-name-in-module
from termcolor import colored

RESULT_PASS = "PASS"  # nosec
RESULT_FAIL = "FAIL"


class ValidationResult(BaseModel):
    """ValidationResult object.

    This object is meant to store the result of a given test along with some contextual
    information about the test itself.
    """

    result: str
    schema_id: str
    instance_name: Optional[str]
    instance_location: Optional[str]
    instance_type: Optional[str]
    instance_hostname: Optional[str]
    source: Any = None
    strict: bool = False

    # if failed
    absolute_path: Optional[List[str]] = []
    message: Optional[str]

    @validator("result")
    def result_must_be_pass_or_fail(cls, var):  # pylint: disable=no-self-argument, no-self-use
        """Validate that result either PASS or FAIL."""
        if var.upper() not in [RESULT_PASS, RESULT_FAIL]:
            raise ValueError("must be either PASS or FAIL")
        return var.upper()

    def passed(self):
        """Return True or False to indicate if the test has passed.

        Returns
            Bool: indicate if the test passed or failed
        """
        if self.result == RESULT_PASS:
            return True

        return False

    def print(self):
        """Print the result of the test in CLI."""
        if self.passed():
            self.print_passed()
        else:
            self.print_failed()

    def print_failed(self):
        """Print the result of the test to CLI when the test failed."""
        # Construct the message dynamically based on the instance_type
        msg = colored("FAIL", "red") + f" | [ERROR] {self.message}"
        if self.instance_type == "FILE":
            msg += f" [{self.instance_type}] {self.instance_location}/{self.instance_name}"

        elif self.instance_type == "HOST":
            msg += f" [{self.instance_type}] {self.instance_hostname}"

        msg += f" [PROPERTY] {':'.join(str(item) for item in self.absolute_path)}"

        # print the msg
        print(msg)

    def print_passed(self):
        """Print the result of the test to CLI when the test passed."""
        if self.instance_type == "FILE":
            print(colored("PASS", "green") + f" | [{self.instance_type}] {self.instance_location}/{self.instance_name}")

        if self.instance_type == "HOST":
            print(
                colored("PASS", "green")
                + f" | [{self.instance_type}] {self.instance_hostname} [SCHEMA ID] {self.schema_id}"
            )



"""Schema manager."""
import os
import sys
import json

import jsonref
from termcolor import colored
from rich.console import Console
from rich.table import Table

from schema_enforcer.utils import load_file, find_file, find_files, dump_data_to_yaml
from schema_enforcer.validation import ValidationResult, RESULT_PASS, RESULT_FAIL
from schema_enforcer.exceptions import SchemaNotDefined, InvalidJSONSchema
from schema_enforcer.utils import error, warn
from schema_enforcer.schemas.jsonschema import JsonSchema
from schema_enforcer.schemas.validator import load_validators


class SchemaManager:
    """The SchemaManager class is designed to load and organaized all the schemas."""

    def __init__(self, config):
        """Initialize the SchemaManager and search for all schema files in the schema_directories.

        Args:
            config (Config): Instance of Config object returned by schema_enforcer.config.load() method.
        """
        self.schemas = {}
        self.config = config

        full_schema_dir = f"{config.main_directory}/{config.schema_directory}/"

        files = find_files(
            file_extensions=[".yaml", ".yml", ".json"],
            search_directories=[full_schema_dir],
            excluded_filenames=config.schema_file_exclude_filenames,
            return_dir=True,
        )

        # For each schema file, determine the absolute path to the directory
        # Create and save a JsonSchema object for each file
        for root, filename in files:
            root = os.path.realpath(root)
            schema = self.create_schema_from_file(root, filename)
            self.schemas[schema.get_id()] = schema

        # Load validators
        validators = load_validators(config.validator_directory)
        self.schemas.update(validators)

    def create_schema_from_file(self, root, filename):  # pylint: disable=no-self-use
        """Create a new JsonSchema object for a given file.

        Load the content from disk and resolve all JSONRef within the schema file.

        Args:
            root (string): Absolute location of the file in the filesystem.
            filename (string): Name of the file.

        Returns:
            JsonSchema: JsonSchema object newly created.
        """
        file_data = load_file(os.path.join(root, filename))

        # TODO Find the type of Schema based on the Type, currently only jsonschema is supported
        # schema_type = "jsonschema"
        base_uri = f"file:{root}/"
        schema_full = jsonref.JsonRef.replace_refs(file_data, base_uri=base_uri, jsonschema=True, loader=load_file)
        schema = JsonSchema(schema=schema_full, filename=filename, root=root)
        # Only add valid jsonschema files and raise an exception if an invalid file is found
        valid = all((result.passed() for result in schema.check_if_valid()))
        if not valid:
            raise InvalidJSONSchema(schema)
        return schema

    def iter_schemas(self):
        """Return an iterator of all schemas in the SchemaManager.

        Returns:
            Iterator: Iterator of all schemas in K,v format (key, value).
        """
        return self.schemas.items()

    def print_schemas_list(self):
        """Print the list of all schemas to the cli.

        To avoid very long location string, dynamically replace the current dir with a dot.
        """
        console = Console()
        table = Table(show_header=True, header_style="bold cyan")
        current_dir = os.getcwd()
        table.add_column("Schema ID", style="bright_green")
        table.add_column("Type")
        table.add_column("Location")
        table.add_column("Filename")
        for schema_id, schema in self.iter_schemas():
            table.add_row(schema_id, schema.schematype, schema.root.replace(current_dir, "."), schema.filename)
        console.print(table)

    def dump_schema(self, schema_id=None):
        """Dump schema with references resolved.

        Dumps all schemas or a single schema represented by schema_id.

        Args:
            schema_id (str): The unique identifier of a schema.

        Returns: None
        """
        if schema_id:
            schema = self.schemas.get(schema_id, None)
            if schema is None:
                raise ValueError(f"Could not find schema ID {schema_id}")
            print(json.dumps(schema.data, indent=2))
        else:
            for _, schema in self.iter_schemas():
                print(json.dumps(schema.data, indent=2))

    def test_schemas(self):
        """Validate all schemas pass the tests defined for them.

        For each schema, 3 set of tests will be potentially executed.
          - schema must be Draft7 valid.
          - Valid tests must pass.
          - Invalid tests must pass.
        """
        error_exists = False

        for schema_id, schema in self.iter_schemas():

            schema_valid = schema.check_if_valid()
            valid_results = self.test_schema_valid(schema_id)
            invalid_results = self.test_schema_invalid(schema_id)

            for result in schema_valid + valid_results + invalid_results:
                if not result.passed():
                    error_exists = True

                result.print()

        if not error_exists:
            print(colored("ALL SCHEMAS ARE VALID", "green"))

    def test_schema_valid(self, schema_id, strict=False):
        """Execute all valid tests for a given schema.

        Args:
            schema_id (str): The unique identifier of a schema.

        Returns:
            list of ValidationResult.
        """
        schema = self.schemas[schema_id]

        valid_test_dir = self._get_test_dir_absolute(test_type="valid", schema_id=schema_id)

        valid_files = find_files(
            file_extensions=[".yaml", ".yml", ".json"],
            search_directories=[valid_test_dir],
            excluded_filenames=[],
            return_dir=True,
        )

        results = []

        for root, filename in valid_files:

            test_data = load_file(os.path.join(root, filename))

            for result in schema.validate(test_data, strict=strict):
                result.instance_name = filename
                result.instance_location = root
                result.instance_type = "TEST"
                results.append(result)

        return results

    def test_schema_invalid(self, schema_id):  # pylint: disable=too-many-locals
        """Execute all invalid tests for a given schema.

        - Acquire structured data to be validated against a given schema. Do this by searching for a file named
          "data.yml", "data.yaml", or "data.json" within a directory of hierarchy
          "./<test_directory>/<schema_id>/invalid/"
        - Acquire results expected after data is validated against a giveh schema. Do this by searching for a file
          named "results.yml", "results.yaml", or "results.json" within a directory of hierarchy
          "./<test_directory>/<schema_id>/results/"
        - Validate expected results match actual results of data after it is checked for adherence to schema.

        Args:
            schema_id (str): The unique identifier of a schema.

        Returns:
            list of ValidationResult objects.
        """
        schema = self.schemas.get(schema_id, None)

        if schema is None:
            raise ValueError(f"Could not find schema ID {schema_id}")

        invalid_test_dir = self._get_test_dir_absolute(test_type="invalid", schema_id=schema_id)
        test_dirs = next(os.walk(invalid_test_dir))[1]

        results = []
        for test_dir in test_dirs:
            schema.clear_results()
            data_file_path = os.path.join(invalid_test_dir, test_dir, "data")
            data_file = find_file(data_file_path)
            expected_results_file_path = os.path.join(invalid_test_dir, test_dir, "results")
            expected_results_file = find_file(expected_results_file_path)

            if not data_file:
                warn(f"Could not find data file {data_file_path}. Skipping...")
                continue

            if not expected_results_file:
                warn(f"Could not find expected_results_file {expected_results_file_path}. Skipping...")
                continue

            data = load_file(data_file)
            expected_results = load_file(expected_results_file)

            tmp_results = schema.validate_to_dict(data)

            # Currently the expected results are using OrderedDict instead of Dict
            # the easiest way to remove that is to dump into JSON and convert back into a "normal" dict
            expected_results = json.loads(json.dumps(expected_results["results"]))

            results_sorted = sorted(tmp_results, key=lambda i: i.get("message", ""))
            expected_results_sorted = sorted(expected_results, key=lambda i: i.get("message", ""))

            params = dict(
                schema_id=schema_id, instance_type="TEST", instance_name=test_dir, instance_location=invalid_test_dir
            )
            if results_sorted != expected_results_sorted:
                params["result"] = RESULT_FAIL
                params[
                    "message"
                ] = f"Invalid test results do not match expected test results from {expected_results_file}"
            else:
                params["result"] = RESULT_PASS

            val = ValidationResult(**params)
            results.append(val)

        return results  # [ ValidationResult(**result) for result in results ]

    def generate_invalid_tests_expected(self, schema_id):
        """Generate expected invalid test results for a given schema.

        Args:
            schema_id (str): unique identifier of a schema
        """
        # TODO: Refactor this into a method. Exists in multiple places
        schema = self.schemas.get(schema_id, None)
        if schema is None:
            raise ValueError(f"Could not find schema ID {schema_id}")

        invalid_test_dir = self._get_test_dir_absolute(test_type="invalid", schema_id=schema_id)
        test_dirs = next(os.walk(invalid_test_dir))[1]

        # For each test, load the data file, test the data against the schema and save the results
        for test_dir in test_dirs:
            schema.clear_results()
            data_file_path = os.path.join(invalid_test_dir, test_dir, "data")
            data_file = find_file(data_file_path)

            if not data_file:
                warn(f"Could not find data file {data_file_path}")

            data = load_file(data_file)
            results = schema.validate_to_dict(data)
            self._ensure_results_invalid(results, data_file)

            result_file = os.path.join(invalid_test_dir, test_dir, "results.yml")
            dump_data_to_yaml({"results": results}, result_file)
            print(f"Generated/Updated results file: {result_file}")

    def validate_schemas_exist(self, schema_ids):
        """Validate that each schema ID in a list of schema IDs exists.

        Args:
            schema_ids (list): A list of schema IDs, each of which should exist as a schema object.
        """
        if not isinstance(schema_ids, list):
            raise TypeError("schema_ids argument passed into validate_schemas_exist must be of type list")
        for schema_id in schema_ids:
            if not self.schemas.get(schema_id, None):
                raise SchemaNotDefined(f"Schema ID {schema_id} declared but not defined")

    @property
    def test_directory(self):
        """Return the path to the main schema test directory."""
        return os.path.join(self.config.main_directory, self.config.test_directory)

    def _get_test_dir_absolute(self, test_type, schema_id):
        """Get absolute path of directory in which schema unit tests exist.

        Args:
            test_type (str): Test type. One of "valid" or "invalid"
            schema_id (str): Schema ID for which to get test dir absolute path

        Returns:
            str: Full path of test directory.
        """
        if test_type not in ["valid", "invalid"]:
            raise ValueError(f"Test type parameter was {test_type}. Must be one of 'valid' or 'invalid'")

        if not self.schemas.get(schema_id, None):
            raise ValueError(f"Could not find schema ID {schema_id}")

        root = os.path.abspath(os.getcwd())
        short_schema_id = schema_id.split("/")[1] if "/" in schema_id else schema_id
        test_dir = os.path.join(root, self.test_directory, short_schema_id, test_type)

        if not os.path.exists(test_dir):
            error(f"Tried to search {test_dir} for {test_type} data, but the path does not exist.")
            sys.exit(1)

        return test_dir

    @staticmethod
    def _ensure_results_invalid(results, data_file):
        """Ensures each result is schema valid in a list of results data structures.

        Args:
            results(dict): List of Dicts of results. Each result dict must include a 'result' key of 'PASS' or 'FAIL'
            data_file (str): Data file which should be schema invalid

        Raises:
            error: Raises an error and calls sys.exit(1) if one of the results objects is schema valid.
        """
        results_pass_or_fail = [result["result"] for result in results]

        if "PASS" in results_pass_or_fail:
            error(f"{data_file} is schema valid, but should be schema invalid as it defines an invalid test")
            sys.exit(1)



"""class to manage jsonschema type schema."""
import copy
import pkgutil
import json

from jsonschema import Draft7Validator, draft7_format_checker  # pylint: disable=import-self
from schema_enforcer.schemas.validator import BaseValidation
from schema_enforcer.validation import ValidationResult, RESULT_FAIL, RESULT_PASS

# TODO do we need to catch a possible exception here ?
v7data = pkgutil.get_data("jsonschema", "schemas/draft7.json")
v7schema = json.loads(v7data.decode("utf-8"))


class JsonSchema(BaseValidation):  # pylint: disable=too-many-instance-attributes
    """class to manage jsonschema type schemas."""

    schematype = "jsonchema"

    def __init__(self, schema, filename, root):
        """Initilize a new JsonSchema object from a dict.

        Args:
            schema (dict): Data representing the schema. Must be jsonschema valid.
            filename (string): Name of the schema file on the filesystem.
            root (string): Absolute path to the directory where the schema file is located.
        """
        super().__init__()
        self.filename = filename
        self.root = root
        self.data = schema
        self.id = self.data.get("$id")  # pylint: disable=invalid-name
        self.top_level_properties = set(self.data.get("properties"))
        self.validator = None
        self.strict_validator = None
        self.format_checker = draft7_format_checker

    def get_id(self):
        """Return the unique ID of the schema."""
        return self.id

    def validate(self, data, strict=False):
        """Validate a given data with this schema.

        Args:
            data (dict, list): Data to validate against the schema.
            strict (bool, optional): if True the validation will automatically flag additional properties. Defaults to False.

        Returns:
            Iterator: Iterator of ValidationResult
        """
        if strict:
            validator = self.__get_strict_validator()
        else:
            validator = self.__get_validator()

        has_error = False
        for err in validator.iter_errors(data):

            has_error = True
            self.add_validation_error(err.message, absolute_path=list(err.absolute_path))

        if not has_error:
            self.add_validation_pass()
        return self.get_results()

    def validate_to_dict(self, data, strict=False):
        """Return a list of ValidationResult objects.

        These are generated with the validate() function in dict() format instead of as a Python Object.

        Args:
            data (dict, list): Data to validate against the schema.
            strict (bool, optional): if True the validation will automatically flag additional properties. Defaults to False.

        Returns:
            list of dictionnaries containing the results.
        """
        return [
            result.dict(exclude_unset=True, exclude_none=True) for result in self.validate(data=data, strict=strict)
        ]

    def __get_validator(self):
        """Return the validator for this schema, create if it doesn't exist already.

        Returns:
            Draft7Validator: The validator for this schema.
        """
        if self.validator:
            return self.validator

        self.validator = Draft7Validator(self.data, format_checker=self.format_checker)

        return self.validator

    def __get_strict_validator(self):
        """Return a strict version of the Validator, create it if it doesn't exist already.

        To create a strict version of the schema, this function adds `additionalProperties` to all objects in the schema.

        Returns:
            Draft7Validator: Validator for this schema in strict mode.
        """
        # TODO Currently the function is only modifying the top level object, need to add that to all objects recursively
        if self.strict_validator:
            return self.strict_validator

        # Create a copy if the schema first and modify it to insert `additionalProperties`
        schema = copy.deepcopy(self.data)

        if schema.get("additionalProperties", False) is not False:
            print(f"{schema['$id']}: Overriding existing additionalProperties: {schema['additionalProperties']}")

        schema["additionalProperties"] = False

        # TODO This should be recursive, e.g. all sub-objects, currently it only goes one level deep, look in jsonschema for utilitiies
        for prop_name, prop in schema.get("properties", {}).items():
            items = prop.get("items", {})
            if items.get("type") == "object":
                if items.get("additionalProperties", False) is not False:
                    print(
                        f"{schema['$id']}: Overriding item {prop_name}.additionalProperties: {items['additionalProperties']}"
                    )
                items["additionalProperties"] = False

        self.strict_validator = Draft7Validator(schema, format_checker=self.format_checker)
        return self.strict_validator

    def check_if_valid(self):
        """Check if the schema definition is valid against JsonSchema draft7.

        Returns:
            List[ValidationResult]: A list of validation result objects.
        """
        validator = Draft7Validator(v7schema, format_checker=self.format_checker)

        results = []
        has_error = False
        for err in validator.iter_errors(self.data):

            has_error = True

            results.append(
                ValidationResult(
                    schema_id=self.id,
                    result=RESULT_FAIL,
                    message=err.message,
                    absolute_path=list(err.absolute_path),
                    instance_type="SCHEMA",
                    instance_name=self.id,
                    instance_location="",
                )
            )

        if not has_error:
            results.append(
                ValidationResult(
                    schema_id=self.id,
                    result=RESULT_PASS,
                    instance_type="SCHEMA",
                    instance_name=self.id,
                    instance_location="",
                )
            )

        return results



"""Exception classes used in Schema Enforcer.

Copyright (c) 2020 Network To Code, LLC <info@networktocode.com>
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class SchemaNotDefined(Exception):
    """Raised when a schema is declared but not defined.

    Args (Exception): Base Exception Object
    """


class InvalidJSONSchema(Exception):
    """Raised when a JSONschema file is invalid.

    Args (Exception): Base Exception Object
    """

    def __init__(self, schema):
        """Provide instance variables when invalid schema is detected."""
        super().__init__(schema)
        self.schema = schema

    def __str__(self):
        """Generate error string including validation errors."""
        errors = [result.message for result in self.schema.check_if_valid() if not result.passed()]
        message = f"Invalid JSONschema file: {self.schema.filename} - {errors}"
        return message
